#include <stdint.h>
#include <stdio.h>
#include <pru_cfg.h>
#include <pru_intc.h>
#include <sys_tscAdcSs.h>
#include <rsc_types.h>
#include <pru_rpmsg.h>
#include "resource_table.h"
#include "intc_map_0.h"
#include "common.h"

volatile register uint32_t __R31;

/* Host-0 Interrupt sets bit 30 in register R31 */
#define HOST_INT			((uint32_t) 1 << 30)

/* 
 * The PRU-ICSS system events used for RPMsg are defined in the Linux devicetree
 * PRU0 uses system event 16 (To ARM) and 17 (From ARM)
 * PRU1 uses system event 18 (To ARM) and 19 (From ARM)
 */
#define TO_ARM_HOST			16
#define FROM_ARM_HOST			17

/*
 * Using the name 'rpmsg-pru' will probe the rpmsg_pru driver found
 * at linux-x.y.z/drivers/rpmsg/rpmsg_pru.c
 */
#define CHAN_NAME			"rpmsg-pru"
#define CHAN_DESC			"Channel 30"
#define CHAN_PORT			30

/*
 * Used to make sure the Linux drivers are ready for RPMsg communication
 * Found at linux-x.y.z/include/uapi/linux/virtio_config.h
 */
#define VIRTIO_CONFIG_S_DRIVER_OK	4

/* Control Module registers to enable the ADC peripheral */
#define CM_WKUP_CLKSTCTRL  (*((volatile unsigned int *)0x44E00400))
#define CM_WKUP_ADC_TSC_CLKCTRL  (*((volatile unsigned int *)0x44E004BC))

/* payload receives RPMsg message */
char payload[RPMSG_MESSAGE_SIZE];

void init_adc2(){
	while (!(CM_WKUP_ADC_TSC_CLKCTRL == 0x02)) {
		CM_WKUP_CLKSTCTRL = 0;
		CM_WKUP_ADC_TSC_CLKCTRL = 0x02;
	}
    // Disable TSC_ADC_SS for programing
	ADC_TSC.CTRL_bit.ENABLE = 0;
	ADC_TSC.CTRL_bit.STEPCONFIG_WRITEPROTECT_N_ACTIVE_LOW = 1;

    // STEPCONFIG
	ADC_TSC.STEPCONFIG1_bit.MODE = 1;               // continuous
	ADC_TSC.STEPCONFIG1_bit.AVERAGING = 0;          // no avg
	ADC_TSC.STEPCONFIG1_bit.SEL_INP_SWC_3_0 = 7;    // (internal 1/2 Vcc)
	ADC_TSC.STEPCONFIG1_bit.SEL_INM_SWC_3_0 = 8;    // ADCREFM
	ADC_TSC.STEPCONFIG1_bit.FIFO_SELECT = 0;

	ADC_TSC.STEPCONFIG2_bit.MODE = 1;               // continuous
	ADC_TSC.STEPCONFIG2_bit.AVERAGING = 0;          // no avg
	ADC_TSC.STEPCONFIG2_bit.SEL_INP_SWC_3_0 = 1;    // AIN1
	ADC_TSC.STEPCONFIG2_bit.SEL_INM_SWC_3_0 = 8;    // ADCREFM
	ADC_TSC.STEPCONFIG2_bit.FIFO_SELECT = 0;

	ADC_TSC.STEPCONFIG3_bit.MODE = 1;               // continuous
	ADC_TSC.STEPCONFIG3_bit.AVERAGING = 0;          // no avg
	ADC_TSC.STEPCONFIG3_bit.SEL_INP_SWC_3_0 = 2;    // AIN3
	ADC_TSC.STEPCONFIG3_bit.SEL_INM_SWC_3_0 = 8;    // ADCREFM
	ADC_TSC.STEPCONFIG3_bit.FIFO_SELECT = 1;

    // Enable TSC_ADC_SS again
	ADC_TSC.CTRL_bit.STEPCONFIG_WRITEPROTECT_N_ACTIVE_LOW = 0;
	ADC_TSC.CTRL_bit.STEP_ID_TAG = 1;
	ADC_TSC.CTRL_bit.ENABLE = 1;

    // clear errors
	ADC_TSC.IRQSTATUS = 0xffff;
}

uint16_t read_adc2(struct shared_struct *dat, uint8_t step){
    const static uint8_t data_count = 128;

    // clear FIFO buffer
    while(ADC_TSC.FIFO0COUNT){
        dat->voltage[0] = ADC_TSC.FIFO0DATA;
    }
    
	ADC_TSC.STEPENABLE = 1<<step;
    uint8_t i;
    for(i=0; i<data_count; i++){
        while(ADC_TSC.FIFO0COUNT < 2);
        dat->voltage[i] = ADC_TSC.FIFO0DATA;
    }
	ADC_TSC.STEPENABLE = 0x0;
    return(ADC_TSC.IRQSTATUS_RAW);
}

void main(void)
{
	struct pru_rpmsg_transport transport;
	uint16_t src, dst, len;
	volatile uint8_t *status;
	struct shared_struct dat;

	/* 
	 * Allow OCP master port access by the PRU so the PRU can read 
	 * external memories 
	 */
	CT_CFG.SYSCFG_bit.STANDBY_INIT = 0;

	init_adc2();

	/* 
	 * Clear the status of the PRU-ICSS system event that the ARM will 
	 * use to 'kick' us
	 */
	CT_INTC.SICR_bit.STS_CLR_IDX = FROM_ARM_HOST;

	/* Make sure the Linux drivers are ready for RPMsg communication */
	status = &resourceTable.rpmsg_vdev.status;
	while (!(*status & VIRTIO_CONFIG_S_DRIVER_OK)) {
		/* Optional: implement timeout logic */
	};

	/* Initialize the RPMsg transport structure */
	pru_rpmsg_init(&transport, &resourceTable.rpmsg_vring0,
		&resourceTable.rpmsg_vring1, TO_ARM_HOST, FROM_ARM_HOST);

	/* 
	 * Create the RPMsg channel between the PRU and ARM user space using 
	 * the transport structure. 
	 */
	while (pru_rpmsg_channel(RPMSG_NS_CREATE, &transport, CHAN_NAME,
			CHAN_PORT) != PRU_RPMSG_SUCCESS) {
		/* Optional: implement timeout logic */
	};

    while(1){
		// Check register R31 bit 30 to see if the ARM has kicked us //
		if (!(__R31 & HOST_INT))
			continue;
		// Clear the event status
		CT_INTC.SICR_bit.STS_CLR_IDX = FROM_ARM_HOST;

		while (pru_rpmsg_receive(&transport, &src, &dst,
				payload, &len) == PRU_RPMSG_SUCCESS) {
            // wait START
			if(*((pru_cmd *)payload) != PRU_START)
                continue;
            dat.irqstatus_raw = read_adc2(&dat, 1);
			pru_rpmsg_send(&transport, dst, src, &dat, sizeof(shared_struct));
        }
    }
}
